'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; /**
                                                                                                                                                                                                                                                                   *  Copyright (c) Facebook, Inc.
                                                                                                                                                                                                                                                                   *  All rights reserved.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  This source code is licensed under the license found in the
                                                                                                                                                                                                                                                                   *  LICENSE file in the root directory of this source tree.
                                                                                                                                                                                                                                                                   *
                                                                                                                                                                                                                                                                   *  
                                                                                                                                                                                                                                                                   */

exports.handleDidOpenOrSaveNotification = handleDidOpenOrSaveNotification;
exports.handleDidChangeNotification = handleDidChangeNotification;
exports.handleDidCloseNotification = handleDidCloseNotification;
exports.handleInitializeRequest = handleInitializeRequest;
exports.handleCompletionRequest = handleCompletionRequest;
exports.handleDefinitionRequest = handleDefinitionRequest;
exports.getQueryAndRange = getQueryAndRange;

var _babylon = require('babylon');

var _babelTraverse = require('babel-traverse');

var _babelTraverse2 = _interopRequireDefault(_babelTraverse);

var _path = require('path');

var _graphqlLanguageServiceConfig = require('graphql-language-service-config');

var _graphqlLanguageServiceInterface = require('graphql-language-service-interface');

var _graphqlLanguageServiceUtils = require('graphql-language-service-utils');

var _vscodeJsonrpc = require('vscode-jsonrpc');

var _vscodeLanguageserver = require('vscode-languageserver');

var _GraphQLCache = require('./GraphQLCache');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var graphQLCache = void 0;
var languageService = void 0;

// Map { uri => { query, range } }


var textDocumentCache = new Map();

function handleDidOpenOrSaveNotification(params) {
  var _this = this;

  var textDocument, uri, contents, diagnostics, cachedDocument;
  return regeneratorRuntime.async(function handleDidOpenOrSaveNotification$(_context2) {
    while (1) {
      switch (_context2.prev = _context2.next) {
        case 0:
          if (!(!params || !params.textDocument)) {
            _context2.next = 2;
            break;
          }

          throw new Error('`textDocument` argument is required.');

        case 2:
          textDocument = params.textDocument;
          uri = textDocument.uri;
          contents = getQueryAndRange(textDocument.text, textDocument.uri);
          diagnostics = [];

          // Create/modify the cached entry if text is provided.
          // Otherwise, try searching the cache to perform diagnostics.

          if (textDocument.text) {
            invalidateCache(textDocument, uri, contents);
          } else {
            cachedDocument = getCachedDocument(uri);

            if (cachedDocument) {
              contents = cachedDocument.contents;
            }
          }

          _context2.next = 9;
          return regeneratorRuntime.awrap(Promise.all(contents.map(function _callee(_ref) {
            var query = _ref.query,
                range = _ref.range;
            return regeneratorRuntime.async(function _callee$(_context) {
              while (1) {
                switch (_context.prev = _context.next) {
                  case 0:
                    _context.t0 = diagnostics.push;
                    _context.t1 = diagnostics;
                    _context.next = 4;
                    return regeneratorRuntime.awrap(provideDiagnosticsMessage(query, uri, range));

                  case 4:
                    _context.t2 = _context.sent;
                    return _context.abrupt('return', _context.t0.apply.call(_context.t0, _context.t1, _context.t2));

                  case 6:
                  case 'end':
                    return _context.stop();
                }
              }
            }, null, _this);
          })));

        case 9:
          return _context2.abrupt('return', { uri: uri, diagnostics: diagnostics });

        case 10:
        case 'end':
          return _context2.stop();
      }
    }
  }, null, this);
}

function handleDidChangeNotification(params) {
  var _this2 = this;

  var textDocument, contentChanges, contentChange, uri, contents, cachedDocument, diagnostics;
  return regeneratorRuntime.async(function handleDidChangeNotification$(_context4) {
    while (1) {
      switch (_context4.prev = _context4.next) {
        case 0:
          if (!(!params || !params.textDocument || !params.contentChanges)) {
            _context4.next = 2;
            break;
          }

          throw new Error('`textDocument` and `contentChanges` arguments are required.');

        case 2:
          textDocument = params.textDocument;
          contentChanges = params.contentChanges;
          contentChange = contentChanges[contentChanges.length - 1];

          // As `contentChanges` is an array and we just want the
          // latest update to the text, grab the last entry from the array.

          uri = textDocument.uri || params.uri;

          // If it's a .js file, try parsing the contents to see if GraphQL queries
          // exist. If not found, delete from the cache.

          contents = getQueryAndRange(contentChange.text, uri);

          // If it's a .graphql file, proceed normally and invalidate the cache.

          invalidateCache(textDocument, uri, contents);

          cachedDocument = getCachedDocument(uri);

          if (cachedDocument) {
            _context4.next = 11;
            break;
          }

          return _context4.abrupt('return', null);

        case 11:

          // Send the diagnostics onChange as well
          diagnostics = [];
          _context4.next = 14;
          return regeneratorRuntime.awrap(Promise.all(contents.map(function _callee2(_ref2) {
            var query = _ref2.query,
                range = _ref2.range;
            return regeneratorRuntime.async(function _callee2$(_context3) {
              while (1) {
                switch (_context3.prev = _context3.next) {
                  case 0:
                    _context3.t0 = diagnostics.push;
                    _context3.t1 = diagnostics;
                    _context3.next = 4;
                    return regeneratorRuntime.awrap(provideDiagnosticsMessage(query, uri, range));

                  case 4:
                    _context3.t2 = _context3.sent;
                    return _context3.abrupt('return', _context3.t0.apply.call(_context3.t0, _context3.t1, _context3.t2));

                  case 6:
                  case 'end':
                    return _context3.stop();
                }
              }
            }, null, _this2);
          })));

        case 14:
          return _context4.abrupt('return', { uri: uri, diagnostics: diagnostics });

        case 15:
        case 'end':
          return _context4.stop();
      }
    }
  }, null, this);
}

function handleDidCloseNotification(params) {
  var textDocument;
  return regeneratorRuntime.async(function handleDidCloseNotification$(_context5) {
    while (1) {
      switch (_context5.prev = _context5.next) {
        case 0:
          if (!(!params || !params.textDocument)) {
            _context5.next = 2;
            break;
          }

          throw new Error('`textDocument` is required.');

        case 2:
          textDocument = params.textDocument;


          if (textDocumentCache.has(textDocument.uri)) {
            textDocumentCache.delete(textDocument.uri);
          }

        case 4:
        case 'end':
          return _context5.stop();
      }
    }
  }, null, this);
}

function handleInitializeRequest(params, token, configDir) {
  var serverCapabilities;
  return regeneratorRuntime.async(function handleInitializeRequest$(_context6) {
    while (1) {
      switch (_context6.prev = _context6.next) {
        case 0:
          if (!(!params || !params.rootPath)) {
            _context6.next = 2;
            break;
          }

          throw new Error('`rootPath` argument is required.');

        case 2:
          _context6.next = 4;
          return regeneratorRuntime.awrap(initialize(configDir ? configDir.trim() : params.rootPath));

        case 4:
          serverCapabilities = _context6.sent;

          if (serverCapabilities) {
            _context6.next = 7;
            break;
          }

          throw new Error('GraphQL Language Server is not initialized.');

        case 7:
          return _context6.abrupt('return', serverCapabilities);

        case 8:
        case 'end':
          return _context6.stop();
      }
    }
  }, null, this);
}

function handleCompletionRequest(params, token) {
  var textDocument, position, cachedDocument, found, query, range, result;
  return regeneratorRuntime.async(function handleCompletionRequest$(_context7) {
    while (1) {
      switch (_context7.prev = _context7.next) {
        case 0:
          if (!(!params || !params.textDocument || !params.position)) {
            _context7.next = 2;
            break;
          }

          throw new Error('`textDocument` argument is required.');

        case 2:
          textDocument = params.textDocument;
          position = params.position;
          cachedDocument = getCachedDocument(textDocument.uri);

          if (cachedDocument) {
            _context7.next = 7;
            break;
          }

          throw new Error(textDocument.uri + ' is not available.');

        case 7:
          found = cachedDocument.contents.find(function (content) {
            var currentRange = content.range;
            if (currentRange && currentRange.containsPosition(position)) {
              return true;
            }
          });

          if (found) {
            _context7.next = 10;
            break;
          }

          throw new Error(textDocument.uri + ' cannot be found from previously opened files.');

        case 10:
          query = found.query, range = found.range;


          if (range) {
            position.line -= range.start.line;
          }
          _context7.next = 14;
          return regeneratorRuntime.awrap(languageService.getAutocompleteSuggestions(query, position, textDocument.uri));

        case 14:
          result = _context7.sent;
          return _context7.abrupt('return', { items: result, isIncomplete: false });

        case 16:
        case 'end':
          return _context7.stop();
      }
    }
  }, null, this);
}

function handleDefinitionRequest(params, token) {
  var textDocument, position, cachedDocument, found, query, range, result, formatted;
  return regeneratorRuntime.async(function handleDefinitionRequest$(_context8) {
    while (1) {
      switch (_context8.prev = _context8.next) {
        case 0:
          if (!(!params || !params.textDocument || !params.position)) {
            _context8.next = 2;
            break;
          }

          throw new Error('`textDocument` and `position` arguments are required.');

        case 2:
          textDocument = params.textDocument;
          position = params.position;
          cachedDocument = getCachedDocument(textDocument.uri);

          if (cachedDocument) {
            _context8.next = 7;
            break;
          }

          throw new Error(textDocument.uri + ' is not available.');

        case 7:
          found = cachedDocument.contents.find(function (content) {
            var currentRange = content.range;
            if (currentRange && currentRange.containsPosition(position)) {
              return true;
            }
          });

          if (found) {
            _context8.next = 10;
            break;
          }

          throw new Error(textDocument.uri + ' cannot be found from previously opened files.');

        case 10:
          query = found.query, range = found.range;

          if (range) {
            position.line -= range.start.line;
          }
          _context8.next = 14;
          return regeneratorRuntime.awrap(languageService.getDefinition(query, position, textDocument.uri));

        case 14:
          result = _context8.sent;
          formatted = result ? result.definitions.map(function (res) {
            var defRange = res.range;
            return {
              // TODO: fix this hack!
              // URI is being misused all over this library - there's a link that
              // defines how an URI should be structured:
              // https://tools.ietf.org/html/rfc3986
              // Remove the below hack once the usage of URI is sorted out in related
              // libraries.
              uri: res.path.indexOf('file://') === 0 ? res.path : 'file://' + res.path,
              range: defRange
            };
          }) : [];
          return _context8.abrupt('return', formatted);

        case 17:
        case 'end':
          return _context8.stop();
      }
    }
  }, null, this);
}

// Check the uri to determine the file type (JavaScript/GraphQL).
// If .js file, either return the parsed query/range or null if GraphQL queries
// are not found.
function getQueryAndRange(text, uri) {
  // Check if the text content includes a GraphQLV query.
  // If the text doesn't include GraphQL queries, do not proceed.
  if ((0, _path.extname)(uri) === '.js') {
    return parseGraphQLQueryFromText(text);
  } else {
    var _query = text;
    if (!_query) {
      return [];
    }
    var lines = _query.split('\n');
    var _range = new _graphqlLanguageServiceUtils.Range(new _graphqlLanguageServiceUtils.Position(0, 0), new _graphqlLanguageServiceUtils.Position(lines.length - 1, lines[lines.length - 1].length - 1));
    return [{ query: _query, range: _range }];
  }
}

function parseGraphQLQueryFromText(text) {
  var ast = (0, _babylon.parse)(text, {
    sourceType: 'module',
    plugins: ['jsx', 'flow', 'doExpressions', 'objectRestSpread', 'classProperties', 'exportExtensions', 'asyncGenerators', 'functionBind', 'functionSent', 'dynamicImport']
  });
  var contents = [];
  (0, _babelTraverse2.default)(ast, {
    TaggedTemplateExpression: function TaggedTemplateExpression(path) {
      var node = path.node;
      var quasi = node.quasi;
      var tag = node.tag;
      if (tag.name === 'graphql' || tag.name === 'graphql.experimental') {
        var loc = path.node.loc;
        var _query2 = quasi.quasis[0].value.raw;
        var _range2 = new _graphqlLanguageServiceUtils.Range(new _graphqlLanguageServiceUtils.Position(loc.start.line - 1, loc.start.column), new _graphqlLanguageServiceUtils.Position(loc.end.line - 1, loc.end.column));
        contents.push({ query: _query2, range: _range2 });
        return;
      }
    }
  });

  return contents;
}

/**
 * Helper functions to perform requested services from client/server.
 */

function initialize(rootPath) {
  var serverCapabilities, configDir;
  return regeneratorRuntime.async(function initialize$(_context9) {
    while (1) {
      switch (_context9.prev = _context9.next) {
        case 0:
          serverCapabilities = {
            capabilities: {
              completionProvider: { resolveProvider: true },
              definitionProvider: true,
              textDocumentSync: 1
            }
          };
          configDir = (0, _graphqlLanguageServiceConfig.findGraphQLConfigDir)(rootPath);

          if (configDir) {
            _context9.next = 4;
            break;
          }

          return _context9.abrupt('return', null);

        case 4:
          _context9.next = 6;
          return regeneratorRuntime.awrap((0, _GraphQLCache.getGraphQLCache)(configDir));

        case 6:
          graphQLCache = _context9.sent;

          languageService = new _graphqlLanguageServiceInterface.GraphQLLanguageService(graphQLCache);

          return _context9.abrupt('return', serverCapabilities);

        case 9:
        case 'end':
          return _context9.stop();
      }
    }
  }, null, this);
}

function provideDiagnosticsMessage(query, uri, range) {
  var results, queryLines, totalLines, lastLineLength, lastCharacterPosition, offset;
  return regeneratorRuntime.async(function provideDiagnosticsMessage$(_context10) {
    while (1) {
      switch (_context10.prev = _context10.next) {
        case 0:
          _context10.next = 2;
          return regeneratorRuntime.awrap(languageService.getDiagnostics(query, uri));

        case 2:
          results = _context10.sent;

          if (results && results.length > 0) {
            queryLines = query.split('\n');
            totalLines = queryLines.length;
            lastLineLength = queryLines[totalLines - 1].length;
            lastCharacterPosition = new _graphqlLanguageServiceUtils.Position(totalLines, lastLineLength);

            results = results.filter(function (diagnostic) {
              return diagnostic.range.end.lessThanOrEqualTo(lastCharacterPosition);
            });

            if (range) {
              offset = range.start;

              results = results.map(function (diagnostic) {
                return _extends({}, diagnostic, {
                  range: new _graphqlLanguageServiceUtils.Range(new _graphqlLanguageServiceUtils.Position(diagnostic.range.start.line + offset.line, diagnostic.range.start.character), new _graphqlLanguageServiceUtils.Position(diagnostic.range.end.line + offset.line, diagnostic.range.end.character))
                });
              });
            }
          }

          return _context10.abrupt('return', results);

        case 5:
        case 'end':
          return _context10.stop();
      }
    }
  }, null, this);
}

function invalidateCache(textDocument, uri, contents) {
  if (!uri) {
    return;
  }

  if (textDocumentCache.has(uri)) {
    var cachedDocument = textDocumentCache.get(uri);
    if (cachedDocument && cachedDocument.version < textDocument.version) {
      // Current server capabilities specify the full sync of the contents.
      // Therefore always overwrite the entire content.
      textDocumentCache.set(uri, {
        version: textDocument.version,
        contents: contents
      });
    }
  } else {
    textDocumentCache.set(uri, {
      version: textDocument.version,
      contents: contents
    });
  }
}

function getCachedDocument(uri) {
  if (textDocumentCache.has(uri)) {
    var cachedDocument = textDocumentCache.get(uri);
    if (cachedDocument) {
      return cachedDocument;
    }
  }

  return null;
}